# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP transport channel serializer (`channel_serializer`).
class TransportChannelSerializer(object):
    # Not set or applicable.
    NULL = 0
    # Use JSON serializer (for dynamically typed app payload).
    JSON = 1
    # Use MsgPack serializer (for dynamically typed app payload).
    MSGPACK = 2
    # Use CBOR serializer (for dynamically typed app payload).
    CBOR = 3
    # Use UBJSON serializer (for dynamically typed app payload).
    UBJSON = 4
    # Raw pass-through of app payload, uninterpreted in any way.
    OPAQUE = 5
    # Use FlatBuffers serialized (statically typed) payload (https://google.github.io/flatbuffers/index.html).
    FLATBUFFERS = 6
    # Use FlexBuffers serialized (dynamically typed) payload (https://google.github.io/flatbuffers/flexbuffers.html).
    FLEXBUFFERS = 7
