# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# Call cancel mode.
class CancelMode(object):
    # Skip running invocation call.
    SKIP = 0
    # Kill running invocation and call.
    KILL = 1
    # Kill running invocation and call without waiting.
    KILLNOWAIT = 2
