# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto


class AnyMessage(object):
    NONE = 0
    Hello = 1
    Welcome = 2
    Abort = 3
    Challenge = 4
    Authenticate = 5
    Goodbye = 6
    Error = 7
    Publish = 8
    Published = 9
    Subscribe = 10
    Subscribed = 11
    Unsubscribe = 12
    Unsubscribed = 13
    Event = 14
    EventReceived = 15
    Call = 16
    Cancel = 17
    Result = 18
    Register = 19
    Registered = 20
    Unregister = 21
    Unregistered = 22
    Invocation = 23
    Interrupt = 24
    Yield = 25
