# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP transport channel framing (`channel_framing`).
class TransportChannelFraming(object):
    # Not set.
    NULL = 0
    # Raw transport itself is inherently message-based already (e.g. FUNCTION or VIRTIO).
    NATIVE = 1
    # Raw transport itself (e.g. TCP) is stream-based and channel framing applied is WebSocket (RFC6455).
    WEBSOCKET = 2
    # Raw transport itself (e.g. TCP) is stream-based and channel framing applied is RawSocket (WAMP).
    RAWSOCKET = 3
