# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# Key Derivation Functions, e.g. WAMP SCRAM uses a password-based key derivation function (KDF) to hash user passwords. WAMP-SCRAM supports both Argon2 and PBKDF2 as the KDF (see: WAMP protocol specification, "SCRAM Algorithms").
class KDF(object):
    # Not set / no TLS channel binding.
    NULL = 0
    # Argon2id variant of Argon2, version 1.3 - `argon2id13`.
    ARGON2ID13 = 1
    # PBKDF2 - `pbkdf2`.
    PBKDF2 = 2
