# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP Payload Passthru (PPT) application payload cipher (NEW!).
class PPTCipher(object):
    # No valid cipher (unfilled)
    NULL = 0
    # Particular combination of Salsa20 and Poly1305 specified in Daniel J. Bernstein, "Cryptography in NaCl" (https://cr.yp.to/highspeed/naclcrypto-20090310.pdf), see NaCl (https://nacl.cr.yp.to/) "crypto_secretbox_xsalsa20poly1305" (aka WAMP "cryptobox").
    XSALSA20POLY1305 = 1
    # Galois/Counter Mode (GCM), see https://en.wikipedia.org/wiki/Galois/Counter_Mode.
    AES256GCM = 2
