# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP Payload Passthru (PPT) scheme (renamed from `Payload`).
class PPTScheme(object):
    # Unset (plain WAMP application payload)
    NONE = 0
    # WAMP-cryptobox end-to-end encrypted application payload
    CRYPTOBOX = 1
    # MQTT passthrough-mode application payload
    MQTT = 2
    # XBR end-to-end encrypted and Ethereum anchored application payload
    XBR = 3
    # Raw pass-through of app payload, uninterpreted in any way.
    OPAQUE = 4
