# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

from autobahn import flatbuffers
from autobahn.flatbuffers.compat import import_numpy

np = import_numpy()


class Abort(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Abort()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsAbort(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Abort
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Abort
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Abort
    def Reason(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Abort
    def Message(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def AbortStart(builder):
    builder.StartObject(3)


def Start(builder):
    AbortStart(builder)


def AbortAddSession(builder, session):
    builder.PrependUint64Slot(0, session, 0)


def AddSession(builder, session):
    AbortAddSession(builder, session)


def AbortAddReason(builder, reason):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(reason), 0
    )


def AddReason(builder, reason):
    AbortAddReason(builder, reason)


def AbortAddMessage(builder, message):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(message), 0
    )


def AddMessage(builder, message):
    AbortAddMessage(builder, message)


def AbortEnd(builder):
    return builder.EndObject()


def End(builder):
    return AbortEnd(builder)
